#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
globato.processors
~~~~~~~~~~~~~~~~~~~

There will be Fetchez hooks here.

:copyright: (c) 2010-2026 Regents of the University of Colorado
:license: MIT, see LICENSE for more details.
"""

from .filters import *
from .transforms import *
from .sinks import *
from .metadata import *
